/*
    A simple program to solve u_xx = 12x^2 on a domain
    from 0 to 1. With u(x=0) = 0; u(x=1) = 1;
*/
#include "tools.h"
#include "comm/comm.h"
#include "system/system.h"
#ifndef TOOLSNOX11
#include "xtools/basex11.h"
#include "xtools/papps/papps.h"
#include "xtools/xlg/xlg.h"
#endif
#include <math.h>

double *DVCreateVector();
void DVdot(), DVnorm(), DVaxpy(), DVaypx(), DVDestroyVector(), DVcopy();

void Mult(N,x,y)
int    *N;
double *x, *y;
{
  int n = *N, i;
  for ( i=1; i<n-1; i++ ) y[i] = 2.0*x[i] - x[i-1] - x[i+1];
  if (PImytid == 0) {
    y[0] = 2.0*x[0] - x[1];
  }
  else {
    PIbsend(1,&y[1],sizeof(double),PImytid-1,MSG_DBL);
    PIbrecv(2,&y[0],sizeof(double),MSG_DBL);
  }
  if (PImytid == PInumtids-1) {
    y[n-1] = 2.0*x[n-1] - x[n-2];
  }
  else {
    PIbsend(2,&y[n-2],sizeof(double),PImytid+1,MSG_DBL);
    PIbrecv(1,&y[n-1],sizeof(double),MSG_DBL);
  }
}

void GSPre(N,x,y)
int    *N;
double *x, *y;
{
  int    nt = *N, n = *N,  i;
  double *zt, *xt = x, *yt = y;
  if (PImytid > 0)          {nt--; xt++; yt++;}
  if (PImytid < PInumtids-1){nt--;}
  
  zt = DVCreateVector(&nt);
  zt[0]    = .5*xt[0];
  for ( i=1; i<nt; i++ )    zt[i] = .5*(xt[i] + zt[i-1]); 
  yt[nt-1] = .5*zt[nt-1];
  for ( i=nt-2; i>=0; i-- ) yt[i] = .5*(zt[i] + yt[i+1]);
  DVDestroyVector(&nt,zt);
  if (PImytid > 0) {
    PIbsend(1,&y[1],sizeof(double),PImytid-1,MSG_DBL);
    PIbrecv(2,&y[0],sizeof(double),MSG_DBL);
  }
  if (PImytid < PInumtids-1) {
    PIbsend(2,&y[n-2],sizeof(double),PImytid+1,MSG_DBL);
    PIbrecv(1,&y[n-1],sizeof(double),MSG_DBL);
  }
}

void JacobiPre(N,x,y)
int    *N;
double *x, *y;
{
  int    n = *N, i;
  for ( i=0; i<n; i++ )  y[i] = x[i];  
}

void Pre(type,N,x,y)
int    *N,type;
double *x, *y;
{
  if (type == 0)      JacobiPre(N,x,y);
  else if (type == 1) GSPre(N,x,y);
}

int worker( argc, argv )
int  argc;
char **argv;
{
  int      N = 100, i, its, n, offset, ptype = 0;
  double   *f, *u, *x, *ustar, tol = 1.e-6,h, pi = 3.14159265358979,dp;
  double   xstart, time;
#ifndef TOOLSNOX11
  XBWindow win1 = 0, win2 = 0, win3 = 0;

  if (!SYArgHasName(&argc,argv,0,"-nox")) {
    win1 = XBOpenPerProcessorWindow("Solution",0);
    win2 = XBOpenPerProcessorWindow("Residual",1); 
    win3 = XBOpenPerProcessorWindow("Error",2); 
  }
#endif

  if (SYArgHasName(&argc,argv,0,"-gs")) ptype = 1;
  SYArgGetInt(&argc,argv,0,"-N",&N);
  h = 1./((double) (N+1)); 
  n = (N + PInumtids - 1)/PInumtids;
  offset = n*PImytid;
  if (n + offset > N) n = N - offset;
  if (PImytid > 0) n++;
  if (PImytid < PInumtids-1) n++;
  if (PImytid == 0) xstart = h;
  else              xstart = h*offset;

  x = DVCreateVector(&n); u     = DVCreateVector(&n);
  f = DVCreateVector(&n); ustar = DVCreateVector(&n);

  /* Form  x and f */
  x[0] = xstart; for ( i=1; i<n; i++ ) x[i] = x[i-1] + h;
  for ( i=0; i<n; i++ ) { 
    f[i]     = -12.*h*h*x[i]*x[i]; 
    u[i]     = 0.0;
    ustar[i] = pow(x[i],4.0);
  }
  /* adjustment for boundary condition */
  if (PImytid == PInumtids-1) f[n-1] += 1.0;

  time = SYGetElapsedTime();
#ifndef TOOLSNOX11
  its = CGSolve(&n,f,u,N,tol,win1,win2,win3,x,ustar,ptype);
#else
  its = CGSolve(&n,f,u,N,tol,(void *)0, (void *)0, (void *)0,x,ustar,ptype);
#endif
  time = SYGetElapsedTime() - time;

  DVaxpy( &n, -1.0, u,ustar ); DVnorm( &n, ustar, &dp);
  if (PImytid == 0) {
    printf("Time %g Number of iterations %d Error %g\n",time,its,sqrt(h)*dp);
  }

  DVDestroyVector(&n,x); DVDestroyVector(&n,u); 
  DVDestroyVector(&n,f); DVDestroyVector(&n,ustar);
#ifndef TOOLSNOX11
  if (win1) {XBWinDestroy(win1); XBWinDestroy(win2); XBWinDestroy(win3);}
#endif
  return 0;
}

#ifndef TOOLSNOX11
void GraphSolution(win,x,u,n)
int      n;
double   *x, *u;
XBWindow *win;
{
  int         i;
  XBLineGraph lg = 0;
  double      umin = 1.e8,umax = 1.e-8,work;
  lg = XBLineGraphCreate(win, 1);
  for ( i=0; i<n; i++ ) {
    if (u[i] > umax)      umax = u[i];
    else if (u[i] < umin) umin = u[i];
  }
  PIgdmax(&umax,1,&work,PSAllProcs);
  PIgdmin(&umin,1,&work,PSAllProcs);
  for ( i=0; i<n; i++ ) {XBLineGraphAddPoint(lg,x+i,u+i);}
  XBLineGraphSetLimits(lg,x[0],x[n-1],umin,umax);
  XBLineGraphDraw(lg);
  XBLineGraphDestroy(lg);
}
#endif

/*                         The Conjugate gradient code */
#include <stdio.h>
#include <math.h>
#include "tools.h"
#ifndef TOOLSNOX11
#include "xtools/basex11.h"
#endif
#include "comm/comm.h"


int  CGSolve(N,B,U,maxit,tol,win1,win2,win3,x,ustar,ptype)
int      *N,maxit,ptype;
double   *B,*U,tol,*x, *ustar;
#ifndef TOOLSNOX11
XBWindow *win1,*win2,*win3;
#else
void *win1, *win2, *win3;
#endif
{
  int       i;
  double    a,beta,betaold,b,dp,*Z,*R,*P, *W;
  
  R       = DVCreateVector(N);                 /* create work vectors */
  Z       = DVCreateVector(N);
  P       = DVCreateVector(N);
  W       = DVCreateVector(N);

  Mult(N,U,R);                                 /*   r <- b - Au      */
  DVaypx(N, -1.0, B, R);
  Pre(ptype,N,R,Z);                            /*     z <- Br        */

  for ( i=0; i<maxit; i++) {
    DVdot( N, R, Z, &beta);                    /*     beta <- r'z    */
    if (i == 0) {
      if (beta == 0.0) break;
      DVcopy( N, Z, P );                       /*     p <- z         */
    }
    else {
      b = beta/betaold;
      DVaypx( N, b, Z, P );                    /*     p <- z + b* p   */
    }
    betaold = beta;
    Mult(N,P,Z);                               /*     z <- Kp         */
    DVdot( N, P, Z, &dp);
    a = beta/dp;                               /*     a = beta/p'z    */
    DVaxpy( N, a, P,U );                       /*     u <- u + ap     */
    DVaxpy( N, -a, Z,R );                      /*     r <- r - az     */
    Pre(ptype,N,R,Z);                          /*     z <- Br         */
    DVnorm( N, Z, &dp);                        /*    dp <- z'*z       */
#ifndef TOOLSNOX11
    if (win1) {
      GraphSolution(win1,x,U,*N);GraphSolution(win2,x,R,*N);
      DVcopy(N,U,W); DVaxpy( N, -1.0, ustar,W );
      GraphSolution(win3,x,W,*N);
      if (PImytid == 0) printf("It %d Residual %g \n",i,dp);
    }
#endif
    if (dp < tol) break;
  }
  DVDestroyVector(N,R);  DVDestroyVector(N,Z); 
  DVDestroyVector(N,P);  DVDestroyVector(N,W);
  return i;
}

/*                      The vector operations */
void DVdot( N, x, y, z )
int             *N;
register double *x, *y, *z;
{
  int    n = *N;
  double sum = 0.0,work;
  if (PImytid > 0) {x++;y++;n--;}
  if (PImytid < PInumtids-1) n--;
  while (n--) sum += (*x++)*(*y++);
  PIgdsum(&sum,1,&work,PSAllProcs);
  *z = sum;
}

void DVnorm( N, x, z )
int             *N;
register double *x, *z;
{
  int    n = *N;
  double sum = 0.0, work;
  if (PImytid > 0) {x++; n--;}
  if (PImytid < PInumtids-1) n--;
  while (n--) {sum += (*x)*(*x); x++;}
  PIgdsum(&sum,1,&work,PSAllProcs);
  *z = sqrt( sum );
}

void DVcopy( N, x, y )
int             *N;
register double *x, *y;
{
  register int n = *N;
  while (n--) *y++ = *x++;
}

void DVaxpy( N, alpha, x, y )
int             *N;
register double alpha, *x, *y;
{
  register int n = *N;
  while (n--) { *y += alpha*(*x++); y++;}
}

void DVaypx( N, alpha, x, y )
int             *N;
register double alpha, *x, *y;
{
  register int n = *N;
  while (n--) { *y = alpha*(*y) + *x++; y++;}
}

double *DVCreateVector( N )
int *N;
{
  return (double *)MALLOC( *N * sizeof(double) );
}

void DVDestroyVector(N, v )
int  *N;
void *v;
{
  FREE(v);
}
