#include "tools.h"
#include "comm/comm.h"
#include "system/system.h"
#include "system/time/usec.h"
/* #MALLOC DECL# */
/*****************************************************************************

  Each collection of test routines contains:

  Initialization function (returns pointer to context to pass back to test
  functions

  Routine to return test function (and set parameters) based on command-line
  arguments

  Routine to output "help" text

 *****************************************************************************/

/*****************************************************************************
 
 Here are the test functions for the collective (global) operations

 This is designed to allow testing of both "native" (vendor-supplied) and 
 Chameleon collective operations.  Note that Chameleon supports subsets of
 processes, which most vendor (non-MPI) systems do not.
 *****************************************************************************/

/* Structure for the test routines */
typedef struct {
    ProcSet *pset;       /* Procset to test over */
    int     src;         /* Source (for scatter) */
    } GOPctx;

#define GETMEM(type,nitem,p) {\
                          p = (type *)MALLOC( nitem * sizeof(type) ) ;\
				  CHKPTRN(p);}
#define SETMEM(nitem,p) {int _i; for (_i=0; _i<nitem; _i++) p[_i] = 0; }

void *GOPInit( argc, argv )
int  *argc;
char **argv;
{
GOPctx *new;
char   psetname[50];
int    psetlen = 50;

new = NEW( GOPctx );   CHKPTRN(new);
new->pset = 0;
new->src  = 0;

#ifdef CHAMELEON_COMM
if (SYArgGetString( &argc, argv, 1, "-pset", psetname, 50 )) {
    int n1, m1, i;
    new->pset = PSCreate( 1 );
    sscanf( psetname, "%d-%d", &n1, &m1 );
    for (i=n1; i<=m1; i++) {
	PSAddMember( new->pset, &i, 1 );
	}
    PSCompile( new->pset );
    }
#else
new->pset = 0;
#endif 
return new;
}

typedef enum { GDOUBLE, GFLOAT, GINT, GCHAR } GDatatype;
typedef enum { 
    GOPSUM, GOPMIN, GOPMAX, GOPSYNC, GOPBCAST, GOPCOL, GOPCOLX } GOperation;

double TestGDSum(), TestGISum(), TestGCol(), TestGColx(), TestGScat(), 
       TestGSync();
double TestGDSumGlob(), TestGISumGlob(), TestGColGlob(), TestGColxGlob(),
       TestGScatGlob(), TestGSyncGlob();

/* Determine the function from the arguments */
double ((*GetGOPFunction( argc, argv, test_name, units )) ())
int *argc;
char **argv;
char *test_name;
char *units;
{
GOperation op    = GOPSYNC;
GDatatype  dtype = GDOUBLE;
int        use_native = 0;
double     (*f)();
int        rrsize;
extern void gsetopL(), gsetopT(), gsetopTP();
extern void gscattersetR();

/* Default choice */
strcpy( test_name, "sync" );

/* Get information on the actual problem to run */
if (SYArgGetInt( argc, argv, 1, "-rr", &rrsize )) {
    PISetRRSize( rrsize );
    }
if (SYArgHasName( argc, argv, 1, "-ring" )) {
    if (PImytid == 0)
	fprintf( stderr, "Ring not yet supported\n" );
    return 0;
    /*    PISetCombFunc( gsetopR ); */
    }
if (SYArgHasName( argc, argv, 1, "-line" )) {
    PISetCombFunc( gsetopL );
    PISetScatterFunc( gscattersetR, (void (*)())0 );
    }
if (SYArgHasName( argc, argv, 1, "-fast" )) {
    PISetNbrRoutines( PIDefFastTree, (void (*)())0, (void (*)())0 );
    PISetCollectiveTree( PISetTreeNodesFast );
    PISetupCollectiveTree();
    }
if (SYArgHasName( argc, argv, 1, "-tree" ))
    PISetCombFunc( gsetopT );

/* Choose the operations */
if (SYArgHasName( argc, argv, 1, "-dsum" )) {
    op    = GOPSUM;
    dtype = GDOUBLE;
    strcpy( test_name, "dsum" );
    strcpy( units, "(doubles)" );
    }
if (SYArgHasName( argc, argv, 1, "-isum" )) {
    op    = GOPSUM;
    dtype = GINT;
    strcpy( test_name, "isum" );
    strcpy( units, "(ints)" );
    }
if (SYArgHasName( argc, argv, 1, "-sync" )) {
    op    = GOPSYNC;
    strcpy( test_name, "sync" );
    }
if (SYArgHasName( argc, argv, 1, "-scatter" ) ||
    SYArgHasName( argc, argv, 1, "-bcast" )) {
    op    = GOPBCAST;
    dtype = GINT;
    strcpy( test_name, "scatter" );
    strcpy( units, "(ints)" );
    }
if (SYArgHasName( argc, argv, 1, "-col" )) {
    op    = GOPCOL;
    dtype = GINT;
    strcpy( test_name, "col" );
    strcpy( units, "(ints)" );
    }
if (SYArgHasName( argc, argv, 1, "-colx" )) {
    op    = GOPCOLX;
    dtype = GINT;
    strcpy( test_name, "colx" );
    strcpy( units, "(ints)" );
    }
if (SYArgHasName( argc, argv, 1, "-colxex" )) {
    extern void gcolxsetEX();
    op    = GOPCOLX;
    dtype = GINT;
    PISetGcolxFunc( gcolxsetEX );
    strcpy( test_name, "colxex" );
    strcpy( units, "(ints)" );
    }
/* #NONMPI_START# */
if (SYArgHasName( argc, argv, 1, "-native" )) {
    use_native = 1;
    }
else {
    strcat( test_name, "-not native" );
    }
/* #NONMPI_END# */

/* Convert operation and dtype to routine */
f = 0;
switch (op) {
    case GOPSUM:
        switch (dtype) {
	    case GDOUBLE: 
	    f = TestGDSum;
#if defined(GDSUMGLOB)
	    if (use_native) f = TestGDSumGlob;
#else 
	    if (use_native) f = 0;
#endif
	    break;
	    case GINT:
	    f = TestGISum;
#if defined(GISUMGLOB)
	    if (use_native) f = TestGISumGlob;
#else 
	    if (use_native) f = 0;
#endif
	    break;
	    }
	break;
    case GOPMIN:
    case GOPMAX:
	f = 0;
	break;
    case GOPCOL:
	f = TestGCol;
#if defined(GCOLGLOB)
	if (use_native) f = TestGColGlob;
#else 
	if (use_native) f = 0;
#endif
	break;
    case GOPCOLX:
	f = TestGColx;
#if defined(GCOLGLOB)
	if (use_native) f = TestGColxGlob;
#else 
	if (use_native) f = 0;
#endif
	break;
    case GOPBCAST:
	f = TestGScat;
#if defined(GSCATTERGLOB)
	if (use_native) f = TestGScatGlob;
#else 
	if (use_native) f = 0;
#endif	
	break;
    case GOPSYNC:
	f = TestGSync;
#if defined(GSYNCGLOB)
	if (use_native) f = TestGSyncGlob;
#else 
	if (use_native) f = 0;
#endif
	break;
    }
return f;
}

PrintGOPHelp()
{
  fprintf( stderr, "\nCollective Tests:\n" );
  fprintf( stderr, "-dsum     : reduction (double precision)\n" );
  fprintf( stderr, "-isum     : reduction (integer)\n" );
  fprintf( stderr, "-sync     : synchronization\n" );
  fprintf( stderr, "-colx     : collect with known sizes\n" );
  fprintf( stderr, 
                "-colxex   : collect with known sizes with exchange alg.\n" );
  fprintf( stderr, "-scatter  : scatter\n" );
  fprintf( stderr, "-bcast    : another name for -scatter\n" );

  fprintf( stderr, "\nCollective test control:\n" );
  fprintf( stderr, 
	  "-pset n-m            : processor set consisting of nodes n to m" );

  fprintf( stderr, "\nCollective communication algorithms:\n" );
  fprintf( stderr, 
   "-rr n   : use a ready-receiver version for messages n bytes or longer\n" );
  fprintf( stderr, 
   "          (many operations do not yet have rr versions)\n" );
  fprintf( stderr, "-ring   : use a ring algorithm\n" );
  fprintf( stderr, "-tree   : use a tree algorithm\n" );
  fprintf( stderr, "-line   : use a linear algorithm\n" );
/* #NONMPI_START# */
  fprintf( stderr, "-native : use the native routines (if available)\n" );
/* #NONMPI_END# */
  fprintf( stderr, "\nOptions for collective algorithms\n" );
  fprintf( stderr, "-pkt len stride cnt : use packets of length len\n" );
  fprintf( stderr, 
         "-fast               : use a (possibly) contention-free tree\n" );
}

/*****************************************************************************
 Here are the actual routines
 *****************************************************************************/
/* First are the Chameleon versions */
double TestGDSum( reps, len, ctx )
int    reps, len;
GOPctx *ctx;
{
int    i;
ProcSet *pset = ctx->pset;
double *lval, *work, time;
SYusc_time_t t0, t1;

GETMEM(double,len,lval);
GETMEM(double,len,work);
SETMEM(len,lval);

PIgsync( pset );
SYusc_clock(&t0);
for (i=0; i<reps; i++) {
    PIgdsum( lval, len, work, pset );
    }
SYusc_clock(&t1);
if (GETERR) SYexitall( "Error in test", 2 );
PIgsync( pset );
time = SYuscDiff( &t0, &t1 );
PIbcast( &time, sizeof(double), PSISROOT(pset), pset, MSG_DBL );
FREE( lval );
FREE( work );
return time;
}

double TestGISum( reps, len, ctx )
int    reps, len;
GOPctx *ctx;
{
int     i;
ProcSet *pset = ctx->pset;
int     *lval, *work;
double  time;
SYusc_time_t t0, t1;

GETMEM(int,len,lval);
GETMEM(int,len,work);
SETMEM(len,lval);

PIgsync( pset );
SYusc_clock(&t0);
for (i=0; i<reps; i++) {
    PIgisum( lval, len, work, pset );
    }
SYusc_clock(&t1);
if (GETERR) SYexitall( "Error in test", 2 );
PIgsync( pset );
time = SYuscDiff( &t0, &t1 );
PIbcast( &time, sizeof(double), PSISROOT(pset), pset, MSG_DBL );
FREE( lval );
FREE( work );
return time;
}

double TestGScat( reps, len, ctx )
int    reps, len;
GOPctx *ctx;
{
int     i;
ProcSet *pset = ctx->pset;
int     *lval;
double  time;
SYusc_time_t t0, t1;

GETMEM(int,len,lval);
SETMEM(len,lval);

PIgsync( pset );
SYusc_clock(&t0);
for (i=0; i<reps; i++) {
    PIbcast( lval, len, PSISROOT(pset), pset, MSG_INT );
    }
SYusc_clock(&t1);
if (GETERR) SYexitall( "Error in test", 2 );
PIgsync( pset );
time = SYuscDiff( &t0, &t1 );
PIbcast( &time, sizeof(double), PSISROOT(pset), pset, MSG_DBL );
FREE( lval );
return time;
}

#ifdef CHAMELEON_COMM
double TestGCol( reps, len, ctx )
int    reps, len;
GOPctx *ctx;
{
int     i, gsize, glen;
ProcSet *pset = ctx->pset;
int     *lval, *gval;
double  time;
SYusc_time_t t0, t1;

gsize = len * PSnumtids(pset);
GETMEM(int,len,lval);
GETMEM(int,gsize,gval);
SETMEM(len,lval);

PIgsync( pset );
SYusc_clock(&t0);
for (i=0; i<reps; i++) {
    PIgcol( lval, len, gval, gsize, &glen, pset, MSG_INT );
    }
SYusc_clock(&t1);
if (GETERR) SYexitall( "Error in test", 2 );
PIgsync( pset );
time = SYuscDiff( &t0, &t1 );
PIbcast( &time, sizeof(double), PSISROOT(pset), pset, MSG_DBL );
FREE( lval );
FREE( gval );
return time;
}

/* Test with data of the same length */
double TestGColx( reps, len, ctx )
int    reps, len;
GOPctx *ctx;
{
int     i, gsize, glen;
ProcSet *pset = ctx->pset;
int     *lval, *gval, *gsizes;
double  time;
int     np, offset;
SYusc_time_t t0, t1;

np    = PSnumtids(pset);
gsize = len * np;
GETMEM(int,len,lval);
GETMEM(int,gsize,gval);
GETMEM(int,np,gsizes);
for (i=0; i<np; i++) 
    gsizes[i] = len * sizeof(int);
/* Set the values so that we can do a correctness check */
offset = PSmytid(pset)*len;
for (i=0; i<len; i++) 
    lval[i] = offset + i;
PIgsync( pset );
SYusc_clock(&t0);
for (i=0; i<reps; i++) {
    PIgcolx( lval, gsizes, gval, pset, MSG_INT ); 
    }
SYusc_clock(&t1);
if (GETERR) SYexitall( "Error in test", 2 );
PIgsync( pset );
time = SYuscDiff( &t0, &t1 );
PIbcast( &time, sizeof(double), PSISROOT(pset), pset, MSG_DBL );
/* Do an error check */
for (i=0; i<gsize; i++) {
    if (gval[i] != i) 
	fprintf( stderr, "[%d] error in gcolx; gval[%d] = %d\n", 
	        PSmytid(pset), i, gval[i] );
    }
FREE( lval );
FREE( gval );
FREE( gsizes );
return time;
}
#else
double TestGColx( reps, len, ctx )
int    reps, len;
GOPctx *ctx;
{
SYexitall( "Error in test", 2 );
}
double TestGCol( reps, len, ctx )
int    reps, len;
GOPctx *ctx;
{
SYexitall( "Error in test", 2 );
}
#endif

double TestGSync( reps, len, ctx )
int    reps, len;
GOPctx *ctx;
{
int     i;
ProcSet *pset = ctx->pset;
double  time;
SYusc_time_t t0, t1;

PIgsync( pset );
SYusc_clock(&t0);
for (i=0; i<reps; i++) {
    PIgsync( pset );
    }
SYusc_clock(&t1);
if (GETERR) SYexitall( "Error in test", 2 );
PIgsync( pset );
time = SYuscDiff( &t0, &t1 );
PIbcast( &time, sizeof(double), PSISROOT(pset), pset, MSG_DBL );
return time;
}

/* Next are the "vendor" versions (when available) */
#ifdef GDSUMGLOB
double TestGDSumGlob( reps, len, ctx )
int    reps, len;
GOPctx *ctx;
{
int    i;
ProcSet *pset = ctx->pset;
double *lval, *work, time;
SYusc_time_t t0, t1;

GETMEM(double,len,lval);
GETMEM(double,len,work);
SETMEM(len,lval);

PIgsync( pset );
SYusc_clock(&t0);
for (i=0; i<reps; i++) {
    GDSUMGLOB( lval, len, work );
    }
SYusc_clock(&t1);
if (GETERR) SYexitall( "Error in test", 2 );
PIgsync( pset );
time = SYuscDiff( &t0, &t1 );
PIbcast( &time, sizeof(double), PSISROOT(pset), pset, MSG_DBL );
FREE( lval );
FREE( work );
return time;
}
#endif

#ifdef GISUMGLOB
double TestGISumGlob( reps, len, ctx )
int    reps, len;
GOPctx *ctx;
{
int     i;
ProcSet *pset = ctx->pset;
int     *lval, *work;
double  time;
SYusc_time_t t0, t1;

GETMEM(int,len,lval);
GETMEM(int,len,work);
SETMEM(len,lval);

PIgsync( pset );
SYusc_clock(&t0);
for (i=0; i<reps; i++) {
    GISUMGLOB( lval, len, work );
    }
SYusc_clock(&t1);
if (GETERR) SYexitall( "Error in test", 2 );
PIgsync( pset );
time = SYuscDiff( &t0, &t1 );
PIbcast( &time, sizeof(double), PSISROOT(pset), pset, MSG_DBL );
FREE( lval );
FREE( work );
return time;
}
#endif

#ifdef GSCATTERGLOB
double TestGScatGlob( reps, len, ctx )
int    reps, len;
GOPctx *ctx;
{
int     i;
ProcSet *pset = ctx->pset;
int     *lval;
double  time;
SYusc_time_t t0, t1;

GETMEM(int,len,lval);
SETMEM(len,lval);

PIgsync( pset );
SYusc_clock(&t0);
for (i=0; i<reps; i++) {
    GSCATTERGLOB( lval, len, PSISROOT(pset), MSG_INT );
    }
SYusc_clock(&t1);
if (GETERR) SYexitall( "Error in test", 2 );
PIgsync( pset );
time = SYuscDiff( &t0, &t1 );
PIbcast( &time, sizeof(double), PSISROOT(pset), pset, MSG_DBL );
FREE( lval );
return time;
}
#endif

#ifdef GCOLGLOB
double TestGColGlob( reps, len, ctx )
int    reps, len;
GOPctx *ctx;
{
int     i, gsize, glen;
ProcSet *pset = ctx->pset;
int     *lval, *gval;
double  time;
SYusc_time_t t0, t1;

gsize = len * PSnumtids(pset);
GETMEM(int,len,lval);
GETMEM(int,gsize,gval);
SETMEM(len,lval);

PIgsync( pset );
SYusc_clock(&t0);
for (i=0; i<reps; i++) {
    GCOLGLOB( lval, len, gval, gsize, &glen, G_INT );
    }
SYusc_clock(&t1);
if (GETERR) SYexitall( "Error in test", 2 );
PIgsync( pset );
time = SYuscDiff( &t0, &t1 );
PIbcast( &time, sizeof(double), PSISROOT(pset), pset, MSG_DBL );
FREE( lval );
FREE( gval );
return time;
}
#endif 

#ifdef GSYNCGLOB
double TestGSyncGlob( reps, len, ctx )
int    reps, len;
GOPctx *ctx;
{
int     i;
ProcSet *pset = ctx->pset;
double  time;
SYusc_time_t t0, t1;

PIgsync( pset );
SYusc_clock(&t0);
for (i=0; i<reps; i++) {
    GSYNCGLOB();
    }
SYusc_clock(&t1);
if (GETERR) SYexitall( "Error in test", 2 );
PIgsync( pset );
time = SYuscDiff( &t0, &t1 );
PIbcast( &time, sizeof(double), PSISROOT(pset), pset, MSG_DBL );
return time;
}
#endif 

#ifdef GCOLXGLOB
double TestGColxGlob( reps, len, ctx )
int    reps, len;
GOPctx *ctx;
{
int     i, gsize, glen;
ProcSet *pset = ctx->pset;
int     *lval, *gval, *gsizes;
double  time;
int     np;
SYusc_time_t t0, t1;

np    = PSnumtids(pset);
gsize = len * np;
GETMEM(int,len,lval);
GETMEM(int,gsize,gval);
GETMEM(int,np,gsizes);
SETMEM(len,lval);
for (i=0; i<np; i++) 
    gsizes[i] = len * sizeof(int);
PIgsync( pset );
SYusc_clock(&t0);
for (i=0; i<reps; i++) {
    GCOLXGLOB( lval, gsizes, gval, MSG_INT );
    }
SYusc_clock(&t1);
if (GETERR) SYexitall( "Error in test", 2 );
PIgsync( pset );
time = SYuscDiff( &t0, &t1 );
PIbcast( &time, sizeof(double), PSISROOT(pset), pset, MSG_DBL );
FREE( lval );
FREE( gval );
FREE( gsizes );
return time;
}
#endif


