/*
    A simple parallel program to solve u_t + u_x = 0 on a periodic domain
    from 0 to 1. With u(x, t = 0) = sin(2*pi*x);
*/
#include "tools.h"
#include "comm/comm.h"
#ifndef TOOLSNOX11
#include "xtools/basex11.h"
#include "xtools/papps/papps.h"
#include "xtools/xlg/xlg.h"
#endif
#include <math.h>
void     UpDateEndPoints(), TimeStep(), GraphSolution();
int worker( argc, argv )
int  argc;
char **argv;
{
  int      N = 100, n, offset, M, i;
  double   *x, *u, *v, h, k, xstart, pi = 3.14159265358979, lambda, tfinal;
  double   err, work;
#ifndef TOOLSNOX11
  XBWindow win = 0;

  if (!SYArgHasName(&argc,argv,0,"-nox")) {
    win = XBOpenPerProcessorWindow("Solution",0);
  }
#endif
  SYArgGetInt(&argc,argv,0,"-N",&N);
  h = 1./((double) N); k = h; lambda = k/h; M = (int) (.5/k);
  n = (N + PInumtids - 1)/PInumtids;
  offset = n*PImytid;
  if (n + offset > N) n = N - offset;
  n += 2;
  xstart = h*(offset - 1);

  x = (double *) MALLOC( n*sizeof(double) ); CHKPTR(x);
  u = (double *) MALLOC( n*sizeof(double) ); CHKPTR(u);
  v = (double *) MALLOC( n*sizeof(double) ); CHKPTR(v);

  /* Form u at initial time */
  x[0] = xstart; for ( i=1; i<n; i++ ) x[i] = x[i-1] + h;
  for ( i=0; i<n; i++ ) { u[i] = sin(2.0*pi*x[i]); }
#ifndef TOOLSNOX11
  if (win) GraphSolution(win,x,u,n);
#endif

  for ( i=0; i<M; i++ ) {
    TimeStep(u,v,n,lambda);
    UpDateEndPoints(v, n); if (win) GraphSolution(win,x,v,n);
    TimeStep(v,u,n,lambda);
    UpDateEndPoints(u, n); if (win) GraphSolution(win,x,u,n);
  }
  /* Determine Error at final time  */
  tfinal = 2*M*k;
  err = 0.0; for ( i=1; i<n-1; i++ ) {
    err += pow(sin(2.0*pi*(x[i] - tfinal)) - u[i],2.0);
  } 
  PIgdsum(&err,1,&work,PSAllProcs); err = sqrt(err);
  if (PImytid == 0) {
    fprintf(stdout,"Total points %d Two norm of error %g\n",N,err);
  }
  FREE(x); FREE(u); FREE(v); if (win) XBWinDestroy(win);
  return 0;
}

void UpDateEndPoints(u, n)
double *u;
int    n;
{
  int left, right;
  right = PImytid + 1; if (right == PInumtids) right = 0;
  left  = PImytid - 1; if (left  == -1) left = PInumtids - 1;

  PIbsend(1,&u[n-2],sizeof(double),right,MSG_DBL);
  PIbrecv(1,&u[0],  sizeof(double),MSG_DBL); 

  PIbsend(2,&u[1],  sizeof(double),left,MSG_DBL);
  PIbrecv(2,&u[n-1],sizeof(double),MSG_DBL); 
}

void TimeStep(u,v,n,lambda)
double *u, *v, lambda;
int    n;
{
  int i;
  for ( i=1; i<n-1; i++ ) {
    v[i] = .5*( (u[i+1] + u[i-1]) - lambda*(u[i+1] - u[i-1]) );
  }
}

#ifndef TOOLSNOX11
void GraphSolution(win,x,u,n)
int      n;
double   *x, *u;
XBWindow win;
{
  int    i;
  static XBLineGraph lg = 0;
  if (lg) XBLineGraphReset(lg); else lg = XBLineGraphCreate(win, 1);
  for ( i=0; i<n; i++ ) { XBLineGraphAddPoint(lg,x+i,u+i); }
  XBLineGraphSetLimits(lg,x[0],x[n-1],-1.0,1.0);
  XBLineGraphDraw(lg);
}
#endif
