#ifndef lint
static char vcid[] = "$Id: piofb.c,v 1.3 1995/02/24 19:56:01 gropp Exp $";
#endif

/* 
   This file contains the routines for writing fixed blocks in parallel.
   We assume that each task has access to separate disk system; the writes
   and reads need not be synchronized.

   The algorithm is to allocate and manage a read and a write cache into
   which data may be copied before being written.  There are lots of 
   assumptions here about the lack of alaising of the contents of the file; 
   if another method is used to access the file, bad things may happen.

   In order to improve performance, if possible, data is read or written 
   directly from/to the device.  For more generality, variable sized data
   may be used.

   These routines are also structured so that they may be used by sequential
   applications.

   In addition, these routines are prepared to make use of non-blocking
   IO when available.

   An additional optimization that can be taken is to identify a block of
   read/write operations so that random positionings within a file 

   Early trials showed significant performance improvements for unformatted
   writes (factor of 2 on Suns).

   This interface DOES give us the ability to use non-blocking IO routines 
   since we can manage the read and write buffers directly.  Such an interface
   will need multiple cache buffers so that one buffer can drain or fill as 
   the other is used.

   The same kind of thing can be done in the parallel routines to 
   cache up the contributing messages.

   An additional optimization can be taken if these routines can be 
   told in advance that a certain contiguous range of bytes in the file
   will be written.  This permits "out-of-order" additions to the 
   cache buffer without forcing an intervening write.  A complication comes
   if the cache has to be flushed before it is filled.  Unless the cache is
   first filled (by reading), it may cause incorrect output.

   Note that the block-oriented writer/reader avoids this by dividing output
   into a set of datablocks that are guarenteed to be filled first.
 */

#include <stdio.h>
#include "tools.h"
#include "comm/comm.h"
#include "comm/io/pio.h"

/* #define VERBOSE */
#ifdef VERBOSE
#define DBUG(a) {fprintf( stdout, "[%d] %s\n", PImytid, a );}
#else
#define DBUG(a)
#endif

PIFcache *PIFCacheCreate( rsize, wsize )
int rsize, wsize;
{
PIFcache *new;
int      n;
char     *p;

new = NEW(PIFcache);    CHKPTRN(new);
if (rsize < 0) rsize = PIF_RCACHE_SIZE;
if (wsize < 0) wsize = PIF_WCACHE_SIZE;
new->rlen    = rsize;
new->rsize   = -1;
new->wlen    = wsize;
new->rp	     = 0;
new->wp	     = 0;
new->rb	     = 0;
new->wb	     = 0;
new->rrloc   = 0;
new->wrloc   = 0;
new->spos    = 0;
new->epos    = 0;
new->fpos    = 0;
new->nrflush = 0;
new->nwflush = 0;
new->nrbytes = 0;
new->nwbytes = 0;
new->nseeks  = 0;
n	     = rsize + wsize;
p	     = (char *)MALLOC( n );    CHKPTRN(p);
if (rsize) {
    new->rb = p; 
    p       += rsize;
    }
if (wsize) {
    new->wb = p;
    }
return new;
}

void PIFCacheDestroy( c )
PIFcache *c;
{
if (c->rlen) {
    FREE(c->rb);
    }
else {
    FREE(c->wb);
    }
FREE( c );
}

int PIFiWriteCached( fp, buffer, len )
PIFILE   *fp;
char     *buffer;
int      len;
{
int      m, n, nb, lenout;
PIFcache *c = fp->c;

lenout = 0;
/* Fill up cache; flush it if we fill it up */
while (len > 0 && (c->wp > 0 || len < c->wlen)) {
    m = len;
    if (len > c->wlen - c->wp) m = c->wlen - c->wp;
    memcpy( c->wb + c->wp, buffer, m );
    len    -= m;
    buffer += m;
    c->wp  += m;
    lenout += m;
    if (c->wp == c->wlen) {
	/* Flush cache */
	n	    = PIiDefWrite( fp, c->wb, c->wlen );
	c->nwbytes += n;
	c->nwflush++;
	c->fpos	   += n;
	c->wrloc    = c->fpos;
	c->wp	    = 0;
	}
    }
if (len > 0) {
    /* If we got here, we can write a multiple of blocksize out */
    nb		= len / c->wlen;
    n		= PIiDefWrite( fp, buffer, nb * c->wlen );
    c->nwbytes += n;
    c->nwflush++;
    buffer     += n;
    len	       -= n;
    lenout     += n;
    c->fpos    += n;
    c->wrloc	= c->fpos;
    /* Place remainder into cache */
    if (len > 0) {
	memcpy( c->wb, buffer, len );
	c->wp += len;
	lenout += len;
	}
    }
return lenout;
}

int PIFiReadCached( fp, buffer, len )
PIFILE   *fp;
char     *buffer;
int      len;
{
int      m, n, nb, lenin;
PIFcache *c = fp->c;

lenin  = 0;
/* Get data from cache */
if (c->rsize > 0 && c->rp < c->rsize) {
    m = len;
    if (len > c->rsize - c->rp) m = c->rsize - c->rp;
    memcpy( buffer, c->rb + c->rp, m );
#ifdef VERBOSE
    fprintf( stdout, "[%d] Reading %d of %d from cache at [%d:%d]\n",
	     PImytid, m, len, c->rrloc, c->rrloc + c->rsize - 1 );
#endif
    len    -= m;
    buffer += m;
    c->rp  += m;
    lenin  += m;
    }
if (len > 0) {
    /* We can read some blocks directly into memory (using a multiple of
       blocksize) */
    nb     = len / c->rlen;
    if (nb > 0) {
#ifdef VERBOSE
    fprintf( stdout, "[%d] Reading %d of %d from file\n",
	     PImytid, nb * c->rlen, len );
#endif
	n        = PIiDefRead( fp, buffer, nb * c->rlen );
	c->nrbytes += n;
	c->nrflush++;
	buffer   += n;
	len      -= n;
	lenin    += n;
	c->rrloc = c->fpos;
	c->fpos  += n;
	}
    /* Get the remainder by loading the cache */
    if (len > 0) {
	/* This may read too much; needs to handle EOF (PIiDefRead 
	   fails for n < c->rlen; not the desired behavior) */
#ifdef VERBOSE
	fprintf( stdout, "[%d] Reading %d for %d into cache at %d\n",
		 PImytid, c->rlen, len, c->fpos );
#endif
	n        = PIiDefRead( fp, c->rb, c->rlen );
	c->nrbytes += n;
	c->nrflush++;
	c->rsize = n;
	c->rp    = 0;
	c->rrloc = c->fpos;
	c->fpos  += n;
	if (n < len) len = n;
	memcpy( buffer, c->rb, len );
	c->rp    += len;
	lenin    += len;
	}
    }
return lenin;
}

int PIFiFlushCache( fp )
PIFILE *fp;
{
PIFcache *c = fp->c;
int      n;

if (c->wp > 0) {
    /* use c->wlen when writing fixed sized blocks */
    n	      = PIiDefWrite( fp, c->wb, c->wp );
    c->fpos   += n;
    c->wrloc  = fp->fpos;
    c->wp     = 0;
    c->nwflush++;
    c->nwbytes += n;
    }
}

#ifndef SEEK_SET
#include <unistd.h>
#endif
/* For now, a seek discards any cached data.  This is very suboptimal; 
   At the very least, we should permit repositioning within the cache block. 
   This is a problem for writes unless the cache size exactly matches the
   IO block size (otherwise, we need to read to fill the cache).  We 
   will start by checking for the cases where we are seeking to the current
   location */
int PIFiLseekCached( fp, offset, whence, rw )
PIFILE *fp;
int    offset;
int    whence;
int    rw;         /* 1 for write, 0 for read */
{
PIFcache *c = fp->c;

/* 
printf( "seek to %d, current location is %d\n", offset, c->wrloc + c->wp );
 */
#ifdef FOO
PIFiFlushCache( fp );
c->wrloc = offset;
c->rrloc = offset;
c->fpos  = offset;
c->nseeks++;
return PIiDefLseek( fp, offset, whence, rw );

#else
if (whence == SEEK_SET) {
    if (rw && offset == c->wrloc + c->wp) {
	DBUG("seek already at cur write location");
	return offset;
	}
    /* 
    fprintf( stdout, "moving seek pointer to %d from %d\n", 
	     offset, c->wrloc + c->wp ); 
     */
    /* Flush the write cache */
    PIFiFlushCache( fp );
    /* If we are seeking within the read cache, just move the rp */
    if (!rw) {
#ifdef VERBOSE
	fprintf( stdout, "[%d] cache contains bytes [%d:%d]\n", 
		 PImytid, c->rrloc, c->rrloc + c->rsize - 1 );
#endif
	if (c->rsize > 0 && offset >= c->rrloc && 
	    c->rrloc + c->rsize > offset) {
#ifdef VERBOSE
	    if (offset == c->rrloc) {
		DBUG("seek already at cur read location");
		}
	    else {
		fprintf( stdout, 
			 "[%d] Moving within read cache by %d to %d\n", 
			 PImytid, offset - c->rrloc, offset );
		}
#endif
	    c->rp   = offset - c->rrloc;
	    /* c->fpos refers to the UNDERLYING operations, not the
	       local ones */
	    /* c->fpos = offset; */
	    return offset;
	    }
#ifdef VERBOSE
	fprintf( stdout, "[%d] moving seek pointer to %d from %d\n", 
		 PImytid, offset, c->rrloc + c->rp ); 
#endif
	/* This invalidates the current read cache */
	/* c->rp = c->rsize; */
	}
#ifdef VERBOSE
    fprintf( stdout, "[%d] Moving by explicit seek to %d\n", PImytid, offset );
#endif
    /* This invalidates the current read cache */
    c->rp    = c->rsize;
    c->wrloc = offset;
    c->rrloc = offset;
    c->fpos  = offset;
    c->nseeks++;
    return PIiDefLseek( fp, offset, whence, rw );
    }
else {
    /* Try to move the cache pointers */
#ifdef VERBOSE
    fprintf( stdout, "[%d] Moving by explicit seek to %d with whence = %d\n", 
	     PImytid, offset, whence );
#endif
    PIFiFlushCache( fp );
    c->nseeks++;
    c->wrloc = PIiDefLseek( fp, offset, whence, rw );
    c->rrloc = c->wrloc;
    c->fpos  = c->wrloc;
    return c->wrloc;
    }
#endif
}

/*@
  PIFCacheEnable - Enables the read/write caches

  Input Parameters:
. fp - PIF file pointer
. rsize - read cache size.  If 0, use no caching.  If -1, use a default size
. wsize - write cache size.  If 0, use no caching.  If -1, use a default size

  Notes:
  This routine enables the caching of data for better IO performance.  If 
  a file is being accessed sequentially (a common case), then these routines
  may significantly improve performance.
 @*/
void PIFCacheEnable( fp, rsize, wsize )
PIFILE *fp;
int    rsize, wsize;
{
if (fp->pmode == PIO_AS_PARALLEL ||
    (fp->pmode == PIO_AS_SEQUENTIAL && PSISROOT(fp->procset))) {
    fp->c = PIFCacheCreate( rsize, wsize );
    /* Change to cached readers and writers */
    if (wsize != 0) 
	fp->write = PIFiWriteCached;
    if (rsize != 0)
	fp->read  = PIFiReadCached;
    fp->flush = PIFiFlushCache;
    fp->lseek = PIFiLseekCached;
    }
}

/*
   Get cache flush/write statistics
 */
void PIFCacheData( fp, nsks, nwf, nrf, avwlen, avrlen )
PIFILE *fp;
int    *nsks, *nwf, *nrf, *avwlen, *avrlen;
{
PIFcache *c = fp->c;


if (!c) {
    *nsks   = 0;
    *nwf    = 0;
    *nrf    = 0; 
    *avwlen = 0;
    *avrlen = 0;
    }
else {
    *nsks   = c->nseeks;
    *nwf    = c->nwflush;
    *nrf    = c->nrflush;
    *avwlen = (c->nwflush > 0) ? c->nwbytes / c->nwflush : 0;
    *avrlen = (c->nrflush > 0) ? c->nrbytes / c->nrflush : 0;
    }
}
