
/*
      This file contains routines to make it simple to use standard 
io from parallel programs without ugly code. The two main routines 
are 
    PISprintf() - for a single printf() as if there were a single processor
    PIMprintf() - each processor does a printf and the results are 
                  in the order of the processors.

      There are two sets of these routines, one set in the comm 
directory  and another set in the system directory (for those folks who
do not get the comm directory). They are identical exept that the one in 
the comm library has PARALLEL defined.

   Because these routines use varargs.h it is hard to be completely 
portable, so far they are tested on: sun4, solaris, rs6000, IRIX, NeXT,
hpux, paragon, cm5.

   The ANSI C way of doing variable arguements appears to be to use 
stdarg.h but I have yet to find an manual page that describes 
how to do it that I can actually get compiled so for now I stick 
to varargs.h
*/ 
#include "tools.h"

#define PARALLEL

#if defined(PARALLEL)
#include "comm/comm.h"
#else
typedef void ProcSet;
#endif

#include <stdio.h>
#undef __STRICT_ANSI__
#include <varargs.h>

/*
     PIMprintf - provides a MIMD type of printf. On a sequential machine
         this behaves just like a printf; on a parallel machine 
         each processor prints in order. Since this requires much 
         synchronization it should only be used in non time 
         sensitive parts of the code.

  Input Parameters:
.     procset - the processor subset; use PSAllProcs or 0 for default
.     fmt - the usual printf format statement
.     va_alist - the other arguments to printf
*/
int  PIMprintf(procset, fmt, va_alist)
ProcSet *procset;
char    *fmt; 
va_dcl 
{
  va_list list;
  int flag,i;
#if defined(PARALLEL)
  for (i=0; i<=PSnumtids(procset); i++) {
    if (PIgtoken( procset, i )) {
#endif
      va_start(list);
      flag = vprintf(fmt,list);
      va_end(list);
      fflush(stdout);
#if defined(PARALLEL)
    }
  }
#endif
  return flag;
}
/*
     PISprintf - provides a SIMD type of printf. On a sequential machine
         this behaves just like a printf; on a parallel machine only 
         one processor will actually do a print.

  Input Parameters:
.     procset - the processor subset; use PSAllProcs or 0 for default
.     fmt - the usual printf format statement
.     va_alist - the other arguments to printf
*/
int  PISprintf(procset, fmt, va_alist)
ProcSet *procset;
char    *fmt; 
va_dcl
{
  int flag;
  va_list list;
#if defined(PARALLEL)
  if (procset == PSAllProcs && PImytid > 0) return 0;
  if (procset != PSAllProcs && !PSISROOT(procset)) return 0;
#endif
  va_start(list);
  flag = vprintf(fmt,list);
  va_end(list);
  return flag;
} 
/*
     PISfprintf - provides a SIMD type of fprintf. On a sequential machine
         this behaves just like a printf; on a parallel machine only 
         one processor will actually do a print.

  Input Parameters:
.     procset - the processor subset; use PSAllProcs or 0 for default
.     fd - file pointer; for instance stderr
.     fmt - the usual printf format statement
.     va_alist - the other arguments to printf
*/
int  PISfprintf(procset, fd, fmt, va_alist)
ProcSet *procset;
FILE    *fd;
char    *fmt; 
va_dcl
{
  int flag;
  va_list list;
#if defined(PARALLEL)
  if (procset == PSAllProcs && PImytid > 0) return 0;
  if (procset != PSAllProcs && !PSISROOT(procset)) return 0;
#endif
  va_start(list);
  flag = vfprintf(fd,fmt,list);
  va_end(list);
  return flag;
}

/*
     PISfopen - provides a SIMD type of fopen. On a sequential machine
         this behaves just like a fopen; on a parallel machine only 
         one processor will actually do an open.

  Input Parameters:
.     filename - the name of the file to open
.     flag - the usual flags for fopen

*/
FILE  *PISfopen(procset,filename,flag)
char    *filename,*flag;
ProcSet *procset;
{
#if defined(PARALLEL)
  if (procset == PSAllProcs && PImytid > 0) return 0;
  if (procset != PSAllProcs && !PSISROOT(procset)) return 0;
#endif
  return fopen(filename,flag);
}          
          
/*
     PISfclose - provides a SIMD type of fclose. On a sequential machine
         this behaves just like a fclose; on a parallel machine only 
         one processor will actually do an close.

  Input Parameters:
.  file - the file to close

*/
int PISfclose(procset,file)
FILE    *file;
ProcSet *procset;
{
#if defined(PARALLEL)
  if (procset == PSAllProcs && PImytid > 0) return 0;
  if (procset != PSAllProcs && !PSISROOT(procset)) return 0;
#endif
  return fclose(file);
} 

/*
     PISfflush - provides a SIMD type of fflush. On a sequential machine
         this behaves just like a fflush; on a parallel machine only 
         one processor will actually do a fflush.

  Input Parameters:
.  file - the file to flush

*/
int PISfflush(procset,file)
FILE    *file;
ProcSet *procset;
{
#if defined(PARALLEL)
  if (procset == PSAllProcs && PImytid > 0) return 0;
  if (procset != PSAllProcs && !PSISROOT(procset)) return 0;
#endif
  return fflush(file);
} 
