#ifndef lint
static char vcid[] = "$Id$";
#endif

/*
    This file contains code to process various formatting commands
    that are indicated by a "." in the first column

    Currently, we support
    . (. space)   for arguments
    .N name       Insert the previously defined name block (*N name ... N*)
    .n            new line at here
    .vb           Begin verbatim (i.e., LaTeX verbatim or HTML PRE)
    .ve           End verbatim
    .p filename   Picture (PS files)
    .cb           Begin picture caption
    .ce           End picture caption                   

    Soon to be added:
    .c citeref    Citation reference (eventually, we will be able to
                  convert this into a hot reference)
 */

#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include "tools.h"
#include "doc/doc.h"

/* Handle . name - description (argument definition) */
int ProcessArgFmt( fin, fout, outfcn, lastWasNl )
FILE *fin, *fout;
DocOutput *outfcn;
int  *lastWasNl;
{
int c;

SkipWhite( fin );
OutputArgBegin( fout );

/* Get the name */
while ( (c = DocGetChar( fin )) != EOF) {
    OutputChar( fout, c );
    if (isspace(c)) { *lastWasNl = 1; break; }
    }
OutputArgDefn( fin, fout );
SkipWhite( fin );
while ( (c = DocGetChar( fin )) != EOF) {
    OutputChar( fout, c );
    if (c == '\n') { 
	int ns = 0;
	/* Check for a "continued" description, defined as:
	   next line is neither EMPTY nor starts with . or $ */
	/* Peek at next character */
	c = DocGetChar( fin );
	ungetc( (char)c, fin );
	/* Check for next character a formatting command or an
	   empty line */
	if (c == '.' || c == '$' || c == '\n' ||
	    c == '@' || c == 'M' || c == 'D') {
	    *lastWasNl = 1; break;
	    }
	/* Check for a blank line, looking for lines of all blanks
	   as well as those containing only a newline. */
	if (c == ' ' || c == '\t') {
	    while (c == ' ' || c == '\t') {
		ns++;
		c = DocGetChar( fin );
		}
	    ungetc( (char)c, fin );
	    /* Really should check for <char> * /  */
	    if (c == '\n' || c == '@' || c == 'M' || c == 'D') {
		*lastWasNl = 1; break;
		}
	    }
	}
    }
OutputArgEnd( fout );

}


/* Process a verbatim mode line (either $ ... or .<haven't decided>) */
int ProcessVerbatimFmt( fin, fout, outfcn, lastWasNl )
FILE *fin, *fout;
DocOutput *outfcn;
int *lastWasNl; 
{
int c;	

/* Raw mode output. */
c = DocGetChar( fin );
if (c == ' ' || c == '\t' || c == '\n') {
    OutputLineBreak( *lastWasNl, fout );
    }
else {
    fprintf( fout, ".%c", (char)c );
    }
if (c != '\n') {
    /* Preserve spacing */
    while ((c = DocGetChar( fin )) != EOF && isspace(c)) {
	/* Eventually, we want to handle tabs etc */
	OutputBlank( fout );
	}
    ungetc( (char)c, fin );
    /* SkipWhite( fin ); */
    while ( (c = DocGetChar( fin )) != EOF) {
	OutputChar( fout, c );
	if (c == '\n') { *lastWasNl = 1; break; }
	}
    }
}	

/* Process a .n name include.  The n has been read, the name has not */
int ProcessNameFmt( fin, fout, outfcn, lastWasNl )
FILE *fin, *fout;
DocOutput *outfcn;
int *lastWasNl; 
{
int  c;
char name[256], *p;

/* Get the name */
SkipWhite( fin );
p = name;
while ( (c = fgetc( fin )) != EOF && !isspace(c)) *p++ = c;
*p = 0;

if (!IncludeNameBlock( fout, name )) {
    fprintf( stderr, "Could not find include name (%s)\n", name );
    }

}

/* Process a .vb ... .ve block */
int ProcessVerbatimBlockFmt( fin, fout, outfcn, lastWasNl )
FILE *fin, *fout;
DocOutput *outfcn;
int *lastWasNl; 
{
int  c;
char buf[256];

/* Check for .vb */
if ((c = fgetc( fin )) != 'b') {
    fprintf( stderr, "Invalid verbatim command '.v%c'\n", c );
    return;
    }
SkipLine( fin );
OutputVerbatimBegin( fout );
while (fgets( buf, 256, fin )) {
    if (strncmp( buf, ".ve", 3 ) == 0) break;
    fputs( buf, fout );
    }
OutputVerbatimEnd( fout );
fputs( "\n", fout );
}

/* Process a .cb ... .ce block */
int ProcessCaption( fin, fout, outfcn, lastWasNl )
FILE *fin, *fout;
DocOutput *outfcn;
int *lastWasNl; 
{
int  c;
char buf[256];

/* Check next character */
c = fgetc( fin );
if (c == 'b') OutputStartCaption( fout );
else if (c == 'e') OutputEndCaption( fout );
else {
    fprintf( stderr, "Invalid verbatim command '.v%c'\n", c );
    return;
    }
}

/* Process a .p name command */
int ProcessPicture( fin, fout, outfcn, lastWasNl )
FILE *fin, *fout;
DocOutput *outfcn;
int *lastWasNl; 
{
char buf[256];

/* Get filename */
if (!fgets( buf, 256, fin ))
    fprintf( stderr, "Could not read picture filename!\n" );

OutputPicture( fout, buf );
}

int ProcessDotFmt( fin, fout, outfcn, lastWasNl )
FILE *fin, *fout;
DocOutput *outfcn;
int  *lastWasNl;
{
int c, d;

c = fgetc( fin );
switch (c) {
    case ' ': ungetc( ' ', fin ); 
              return ProcessArgFmt( fin, fout, outfcn, lastWasNl );
    case 'N': return ProcessNameFmt( fin, fout, outfcn, lastWasNl );
    case 'n': OutputLineBreak( d, fout );
              break;
    case 'v': return ProcessVerbatimBlockFmt( fin, fout, outfcn, lastWasNl );
    case 'p': return ProcessPicture( fin, fout, outfcn, lastWasNl );
    case 'c': return ProcessCaption( fin, fout, outfcn, lastWasNl );

    default:
    	      fprintf( stderr, "Unknown dot command %c\n", c );
    }
return 0;    
}
