/*
    Provides a "Safe" read and write to disk including byte swapping
for appropriate machines.
*/

#include "tools.h"
#include <stdio.h>
#if defined(cray) || defined(__MSDOS__) || defined(hpux) || defined(T9000)
#include <errno.h>
#else
extern int errno;
#endif
#if !defined(__MSDOS__)
#include <sys/errno.h>
#endif
#include "system/system.h"

/*@
  SYGetByteOrder - Determine the byte ordering

  Returns:
. 1 - if the LSB is first
. 2 - if the LSB is last
. 0 - otherwise

  Notes:
  This function is an approximation; it has only been tested on a few systems.
  It does to compute the byte ordering by storing a 1 in various bytes
  and looking at the individual bytes.
@*/
int SYGetByteOrder( )
{
int l;
char *b = (char *)&l;

l = 1;
if (b[0] == 1) return 1;
if (b[sizeof(int)-1] == 1) return 2;
return 0;
}

/*@
  SYByteSwapInt - Swap bytes in an integer

  Input Parameters:
. buff - buffer
. n    - number of integers
@*/
void SYByteSwapInt(buff,n)
int *buff,n;
{
  int  i,j,tmp;
  char *ptr1,*ptr2 = (char *) &tmp;
  for ( j=0; j<n; j++ ) {
    ptr1 = (char *) (&buff[j]);
    for (i=0; i<sizeof(int); i++) {
      ptr2[i] = ptr1[sizeof(int)-1-i];
    }
    buff[j] = tmp;
  }
}

/*@
  SYByteSwapShort - Swap bytes in a short

  Input Parameters:
. buff - buffer
. n    - number of shorts
@*/
void SYByteSwapShort(buff,n)
short *buff;
int   n;
{
  int  i,j;
  short tmp;
  char *ptr1,*ptr2 = (char *) &tmp;
  for ( j=0; j<n; j++ ) {
    ptr1 = (char *) (&buff[j]);
    for (i=0; i<sizeof(short); i++) {
      ptr2[i] = ptr1[sizeof(int)-1-i];
    }
    buff[j] = tmp;
  }
}

/*@
  SYByteSwapDouble - Swap bytes in a double

  Input Parameters:
. buff - buffer
. n    - number of doubles
@*/
void SYByteSwapDouble(buff,n)
double *buff;
int    n;
{
  int    i,j;
  double tmp,*ptr3;
  char   *ptr1,*ptr2 = (char *) &tmp;
  for ( j=0; j<n; j++ ) {
    ptr3 = &buff[j];
    ptr1 = (char *) ptr3;
    for (i=0; i<sizeof(double); i++) {
      ptr2[i] = ptr1[sizeof(double)-1-i];
    }
    buff[j] = tmp;
  }
}

/*@
  SYSafeRead - This is a routine that reads data in, checking for errors and
   reading multiple blocks as required. Returns 0 on success, != 0 on
   failure. Reads from a file that has been opened with open() not
   fopen().

  Input Parameters:
. n    - number of bytes
. type - type of data: should be SYINT, SYDOUBLE, SYFLOAT, SYCHAR, SYSHORT
. fd   - the file identifier

  Output Parameters:
.   p - the buffer to store the data in.

@*/
int SYSafeRead( fd, p, n, type )
int  fd, n, type;
char *p;
{
int maxblock, wsize, err;
int ntmp = n;
char *ptmp = p;

maxblock = 65536;
/* Read the data in blocks of 65536 (some systems don't like large reads;
   SunOS is a particular example) */
while (n) {
    wsize = (n < maxblock) ? n : maxblock;
    err = read( fd, p, wsize );
    if (err < 0 && errno == EINTR) continue;
    if (err == 0 && wsize > 0) return 1;
    if (err < 0) {
        SETERRC( 1, "Error in reading from file" );
        return 1;
        }
    n -= err;
    p += err;
    }
#if defined(intelnx)
if (type == SYINT) 
    SYByteSwapInt(ptmp,ntmp/sizeof(int));
else if (type == SYDOUBLE) 
    SYByteSwapDouble(ptmp,ntmp/sizeof(double));
#endif
return 0;
}

/*@
  SYSafeWrite - This is a routine that writes data in, checking for errors and
   writing multiple blocks as required. Returns 0 on success, != 0 on
   failure. Writes to a file that has been opened with open() not
   fopen().

  Input Parameters:
. n    - number of bytes
. type - type of data: should be SYINT, SYDOUBLE, SYFLOAT, SYCHAR, SYSHORT
. p - the buffer to copy the data from.
. istemp - flag indicating if the data may be corrupted by this routine.
. fd - the file identifier

@*/
int SYSafeWrite( fd, p, n, type, istemp )
int  fd, n, type, istemp;
char *p;
{
int err, maxblock, wsize;
int  ntmp  = n;
char *ptmp = p;

maxblock = 65536;
/* Write the data in blocks of 65536 (some systems don't like large writes;
   SunOS is a particular example) */
#if defined(intelnx)
if (type == SYINT) 
    SYByteSwapInt(ptmp,ntmp/sizeof(int));
else if (type == SYDOUBLE) 
    SYByteSwapDouble(ptmp,ntmp/sizeof(double));
#endif
while (n) {
    wsize = (n < maxblock) ? n : maxblock;
    err = write( fd, p, wsize );
    if (err < 0 && errno == EINTR) continue;
    if (err != wsize) {
	SETERRC( 1, "Error in writing to file" );
	break;
	}
    n -= wsize;
    p += wsize;
    }

#if defined(intelnx)
if (!istemp) {
    if (type == SYINT) 
	SYByteSwapInt(ptmp,ntmp/sizeof(int));
    else if (type == SYDOUBLE) 
	SYByteSwapDouble(ptmp,ntmp/sizeof(double));
    }
#endif

return n != 0;
}
